/*
 * Asterisk -- A telephony toolkit for Linux.
 *
 * IVR API Header File
 * 
 * Copyright (C) 2005, Anthony Minessale II.
 *
 * Anthony Minessale <anthmct@yahoo.com>
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License
 */
#include <asterisk/ivr_api.h>

struct ast_ivr_menu {
	char *name;
	char *greeting_sound;
	char *short_greeting_sound;
	char *invalid_sound;
	int max_failures;
	int timeout;
	int inlen;
	unsigned int flags;
	struct ast_ivr_menu_action *actions;
	struct ast_ivr_menu *next;
};


struct ast_ivr_menu_action {
	ast_ivr_menu_action_function_t *function;
	ast_ivr_action_t ivr_action;
	char *arg;
	char *bind;
	struct ast_ivr_menu_action *next;
};

static ast_ivr_menu_t *ast_ivr_find_menu(ast_ivr_menu_t *stack, char *name) {
	ast_ivr_menu_t *ret;
	for(ret = stack; ret ; ret = ret->next) {
		if(!name || !strcmp(ret->name, name))
			break;
	}
	return ret;
}

static int stream_file(struct ast_channel *chan, const char *filename)
{
	int res = 0, moh = 0;

	if (ast_test_flag(chan, AST_FLAG_MOH)) {
		ast_moh_stop(chan);
		moh = 1;
	}

	if (ast_streamfile(chan, filename, chan->language))
		return -1;
	res = ast_waitstream(chan, "");
	if (!res && moh)
		ast_moh_start(chan, NULL);

	return res;
}

static int stream_and_wait(struct ast_channel *chan, const char *filename, int to)
{
    int res = 0;
    if ((res = stream_file(chan, filename)))
        return res;
    else
        return ast_waitfordigit(chan, to);
}

static void ast_ivr_menu_stack(ast_ivr_menu_t **top, ast_ivr_menu_t *bottom) 
{
	ast_ivr_menu_t *ptr;
	for(ptr = *top ; ptr && ptr->next ; ptr = ptr->next);
	if(ptr) {
		ptr->next = bottom;
	} else {
		*top = bottom;
	}

}

int ast_ivr_menu_set_flags(ast_ivr_menu_t *menu, unsigned int flags) 
{
	if(flags)
		menu->flags |= flags;
	return menu->flags;
}

int ast_ivr_menu_clear_flags(ast_ivr_menu_t *menu, unsigned int flags) 
{
	if(flags)
		menu->flags &= ~flags;
	return menu->flags;
}

int ast_ivr_menu_test_flags(ast_ivr_menu_t *menu, unsigned int flags) 
{
	return (menu->flags & flags);
}

ast_ivr_menu_t *ast_ivr_menu_init(ast_ivr_menu_t *main,
								  char *name,
								  char *greeting_sound, 
								  char *short_greeting_sound,
								  char *invalid_sound, 
								  int timeout,
								  int max_failures, 
								  int inlen,
								  unsigned int flags)
	 
{
	ast_ivr_menu_t *menu;

	if((menu = malloc(sizeof(ast_ivr_menu_t)))) {
		memset(menu, 0, sizeof(ast_ivr_menu_t));
		menu->name = name;
		menu->greeting_sound = greeting_sound;
		menu->short_greeting_sound = short_greeting_sound;
		menu->invalid_sound = invalid_sound;
		menu->max_failures = max_failures;
		menu->timeout = timeout;
		menu->flags = flags;
		menu->actions = NULL;
		menu->inlen = inlen;

		if (menu->timeout <= 0)
			menu->timeout = 10000;
		if (menu->inlen <= 0)
			menu->inlen = 1;

		if(main) {
			ast_ivr_menu_stack(&main, menu);
		}
	}
	return menu;
}


void ast_ivr_menu_bind_action(ast_ivr_menu_t *menu, ast_ivr_action_t ivr_action, char *arg, char *bind)
{
	ast_ivr_menu_action_t *action;

	if((action = malloc(sizeof(ast_ivr_menu_action_t)))) {
		memset(action, 0, sizeof(ast_ivr_menu_action_t));
		action->bind = bind;
		action->next = menu->actions;
		action->arg = arg;
		action->ivr_action = ivr_action;
		menu->actions = action;
	}
}


void ast_ivr_menu_bind_function(ast_ivr_menu_t *menu, ast_ivr_menu_action_function_t *function, char *bind)
{
	ast_ivr_menu_action_t *action;

	if((action = malloc(sizeof(ast_ivr_menu_action_t)))) {
		memset(action, 0, sizeof(ast_ivr_menu_action_t));
		action->bind = bind;
		action->function = function;
		action->next = menu->actions;
		menu->actions = action;
	}
}

void ast_ivr_menu_free(ast_ivr_menu_t *menu) 
{
	ast_ivr_menu_action_t *ap, *apt;
	for(ap = menu->actions; ap ; apt = NULL) {
		apt = ap;
		ap = ap->next;
		free(apt);
	}
	free(menu);
	menu = NULL;
}

void ast_ivr_menu_free_stack(ast_ivr_menu_t *stack) 
{
	ast_ivr_menu_t *menu, *tmp;
	for (menu = stack ; menu ; tmp = NULL) {
		tmp = menu;
		menu = menu->next;
		ast_ivr_menu_free(tmp);
	}
}

int ast_ivr_menu_execute(struct ast_channel *chan, ast_ivr_menu_t *stack, char *name, void *obj) 
{
	int reps = 0, res = 0, errs = 0, match = 0, running = 1;
	char *greeting_sound = NULL, *aptr;
	char *buf, arg[512];
	ast_ivr_action_t todo = IVR_ACTION_DIE;
	ast_ivr_menu_action_t *ap;
	ast_ivr_menu_t *menu;
	

	if(!(menu = ast_ivr_find_menu(stack, name))) {
		ast_log(LOG_ERROR, "Invalid Menu!\n");
		return -1;
	}
	
	if(!(buf = alloca(menu->inlen))) {
		ast_log(LOG_ERROR, "No Memory!\n");
		return -1;
	}
	
	if(option_verbose > 2)
		ast_verbose(VERBOSE_PREFIX_3 "Executing IVR menu %s\n", menu->name);

	for (reps = 0 ; (running && res == 0 && errs < menu->max_failures) ; reps++) {
		if (reps > 0 && menu->short_greeting_sound) {
			greeting_sound = menu->short_greeting_sound;
		} else {
			greeting_sound = menu->greeting_sound;
		}

		match = 0;
		aptr = NULL;
		memset(buf, 0, menu->inlen);
		memset(arg, 0, sizeof(arg));
		res = ast_app_getdata(chan, greeting_sound, buf, menu->inlen, menu->timeout);
		
		if(res < 0)
			break;
		
		if(!res && !ast_strlen_zero(buf)) {
			for(ap = menu->actions; ap ; ap = ap->next) {
				if(!strcmp(buf, ap->bind)) {
					char *membuf;

					if(option_verbose > 2)
						ast_verbose(VERBOSE_PREFIX_3 "IVR menu %s matched %s\n", menu->name, buf);
					match++;
					errs = 0;
					if(ap->function) {
						todo = ap->function(menu, arg, sizeof(arg), obj);
						aptr = arg;
					} else {
						todo = ap->ivr_action;
						aptr = ap->arg;
					}

					switch(todo) {
					case IVR_ACTION_DIE:
						res = -1;
						break;
					case IVR_ACTION_PLAYSOUND:
						res = stream_file(chan, aptr);
						break;
					case IVR_ACTION_TRANSFER:
						ast_parseable_goto(chan, aptr);
						running = 0;
						break;
					case IVR_ACTION_EXECMENU:
						reps = -1;
						res = ast_ivr_menu_execute(chan, stack, aptr, obj);
						break;
					case IVR_ACTION_EXECAPP:
						if((membuf = ast_strdupa(aptr))) {
							char *appname = membuf;
							char *appargs;
							if ((appargs = strchr(appname, '|'))) {
								struct ast_app *app;
								*appargs = '\0';
								appargs++;
								if((app = pbx_findapp(appname))) {
									res = pbx_exec(chan, app, appargs, 1);
								}
							}
						} else {
							ast_log(LOG_ERROR, "No Memory!\n");
							res = -1;		
						}
						break;
					case IVR_ACTION_BACK:
						res = running = 0;
						break;
					case IVR_ACTION_TOMAIN:
						ast_ivr_menu_set_flags(stack, IVR_MENU_FLAG_FALLTOMAIN);
						res = -1;
						break;
					default:
						ast_log(LOG_WARNING, "Invalid TODO!\n");
						break;
					}
				}
			}


			if(menu == stack) { /* top level */ 
				if (ast_ivr_menu_test_flags(stack, IVR_MENU_FLAG_FALLTOMAIN)) { /* catch the fallback and recover */
					ast_ivr_menu_clear_flags(stack, IVR_MENU_FLAG_FALLTOMAIN);
					res = 0;
					running = 1;
					continue;
				}
			}
		}
		if(!match) {
			if(option_verbose > 2)
				ast_verbose(VERBOSE_PREFIX_3 "IVR menu %s caught invalid input %s\n", menu->name, buf);
			if(menu->invalid_sound) {
				res = stream_file(chan, menu->invalid_sound);
			}
			errs++;
			if(!res)
				res = ast_safe_sleep(chan, 1000);
		}
		
	}
	return res;
}

